/************************************************************************************\
 * > ikernel/eza/arch/cpu.c                                                         *
 * - Provide some functionality to manipulate with CPU vectors                      *
 * - Work with GDT/IDT/TSS                                                          *
 *                                                                                  *
 * Copyright(c) 2006 Dan Kruchinin (aka asgard)                                     *
 *                                                                                  *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     * 
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
\************************************************************************************/


#include <arch/page.h>
#include <arch/cpu.h>
#include <arch/string.h>
#include <arch/types.h>

static gdt_entry gdt[GDT_SIZE];

void init_cpu_vectors()
{
  gdt_init();
  idt_init();
  /* TODO: init TSS and may be vga */
}

static inline void __load_gdt(tdesc_ptr* gdtp)
{
  __asm__ volatile ("lgdt %0\n\t"
					"movw $0x10, %%cx\n\t"
					"movw %%cx, %%es\n\t"
					"movw %%cx, %%ds\n\t"
					"movw %%cx, %%gs\n\t"
					"movw %%cx, %%fs\n\t"
					"movw %%cx, %%ss\n"
					: : "m" (*gdtp));
}

void gdt_init()
{
  uint8_t aflags = 0;
  tdesc_ptr gdtp;

  gdtp.limit = sizeof(gdt) * GDT_SIZE - 1;
  gdtp.base = (uint32_t)gdt;

  memset(gdt, 0, GDT_SIZE);
  
  /* NULL segment */
  __gdt_fill(gdt, 0, 0, aflags, 0, 0);

  /* kernel code segment */
  aflags = DPL_KERNEL | STF_CODE | STF_PRESENT;
  __gdt_fill(gdt + 1, ADDR_LIMIT, 0, aflags, 1, 1);
  aflags = 0;

  /* kernel data segment */
  aflags = DPL_KERNEL | STF_DATA | STF_PRESENT | STF_PAGE_WRITABLE;
  __gdt_fill(gdt + 2, ADDR_LIMIT, 0, aflags, 1, 1);
  aflags = 0;

  /* user code segment */
  aflags = DPL_USER | STF_CODE | STF_PRESENT;
  __gdt_fill(gdt + 3, ADDR_LIMIT, 0, aflags, 1, 1);
  aflags = 0;

  /* user data segment */
  aflags = DPL_USER | STF_DATA | STF_PRESENT | STF_PAGE_WRITABLE;
  __gdt_fill(gdt + 4, ADDR_LIMIT, 0, aflags, 1, 1);
  aflags = 0;

  __load_gdt(&gdtp);
}

void gdt_set_base(gdt_entry* gdte, const uint32_t base)
{
  gdte->base_low = (base & 0xFFFF);
  gdte->base_middle = (base >> 16) & 0xFF;
  gdte->base_high = (base >> 24) & 0xFF;
}

void gdt_set_limit(gdt_entry* gdte, const uint32_t limit)
{
  gdte->limit = limit & 0xFFFF;
  gdte->seg_limit = (limit >> 16) & 0xF;
}

static void __gdt_fill(gdt_entry* gdte, const uint32_t limit, const uint32_t base, const uint8_t aflags, const int32_t opsize, const int32_t granularity)
{
  gdt_set_base(gdte, base);
  gdt_set_limit(gdte, limit);
  gdte->access = aflags;
  gdte->available = 0;
  gdte->opsize = opsize;
  gdte->granularity = granularity;
}
