/*
 * Virtual file system implementation.
 * The part of the Jari microkernel.
 *
 * Written by Alfeiks Kaanoken <akaanoken@gmail.com>
 * (c) 2002, 2006
 *
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *  
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */

#ifndef __VFS_H__
#define __VFS_H__ 

#include <perm.h>
#include <vfs/types.h>

typedef struct __fnode {
  char type; /*type of the file*/
  char *name;
  int lock; /*locking*/
  /*addressing*/
  unsigned long seg;
  unsigned long offset;
  void *data;
  time_t time;
  perm_t perm;
  /*functions for manipulating with fnode*/
  struct fnode_functions *def_func;
} fnode;

typedef struct __dpoint {
  char *name;
  fnode *node;
  struct __dpoint *parent;
  struct __dpoint **pchild;
  fnode **fchild;
} dpoint;

struct fnode_functions {
  /*basic i/o*/
  int (*create)(fnode *, char *, dpoint *);
  int (*open)(fnode *);
  int (*close)(fnode *);
  int (*read)(fnode *, void *, unsigned long);
  int (*write)(fnode *, void *, unsigned long);
  int (*exec)(fnode *, void *);
  /*file i/o*/
  unsigned long (*lseek)(fnode *, unsigned long);
  int (*mmap)(fnode *, void *);
  int (*unlink)(fnode *);
  int (*rename)(fnode *,void *,dpoint *);
  /*links i/o*/
  int (*readlink)(fnode *,dpoint *);
  int (*link)(fnode *, dpoint *,dpoint *);
  int (*follow_link)(fnode *,dpoint *,dpoint *);
  /*points i/o*/
  int (*mkpnt)(fnode *,dpoint *);
  int (*rmpnt)(fnode *,dpoint *);
  int (*read_pnt)(fnode *,void *,dpoint *);
  /*misc i/o*/
  int (*lock)(fnode *);
  int (*unlock)(fnode *);
  /*device i/o*/
  int (*check_active)(fnode *);
  int (*media_change)(fnode *);
  int (*get_dev_load)(fnode *,void *);
  int (*get_dev_type)(fnode *,void *);
  /*permission i/o*/
  int (*get_perm)(fnode *, perm_t *);
  int (*set_perm)(fnode *, perm_t *);
};

typedef struct __super_fnode {
  /*statistics*/
  int state; /*load avergage status*/
  int lock; /*lock state*/
  int m_type; /*media type*/
  /*specific info*/
  int dev_id; /*server id for low level operations*/
  perm_t perm;
  char *mpoint;
  /*fs operations*/
  struct super_fnode_operations *def_func;
} super_fnode;

struct super_fnode_operations {
  fnode *(*get_fnode)(void *);
  fnode *(*unset_fnode)(fnode *);
  int (*read_fnode)(fnode *);
  int (*write_fnode)(fnode *);
  int (*put_fnode)(fnode *);
  int (*stat_fs)(super_fnode *,void *);
  void (*clear_fnode)(fnode *);
  /*additional*/
};

typedef struct __super_fs {
  int rw; /*read only, write only, read write*/
  int srv_id; /*server of the fs, or the internal routnes*/
  char *name; /*name of the filesystem*/
  char *mpoint; /*mount point*/
  super_fnode* (*read_fnode) (super_fnode *, void *);
  struct __super_fs *next;
} super_fs;

int init_vfs(void);

super_fs *create_super_fs(int rw,int srv_id,char *name,char *mpoint,
			  super_fnode *(*read_fnode) (super_fnode *,void *));
int add_super_fs(super_fs *n);
int del_super_fs(char *mpoint);
super_fnode *lookup_superfs_fnode(char *name);
int flush_superfs(void);

/*basic i/o*/
int vfs_create(char *name,perm_t *p,char type);
int vfs_open(char *name);
int vfs_close(char *name);
int vfs_read(char *name,unsigned long size,void *data);
int vfs_write(char *name, unsigned long size,void *data);
int vfs_exec(char *name, void *data);

/*file i/o*/
unsigned long vfs_lseek(char *name, unsigned long seek);
int vfs_mmap(char *name,void *data);
int vfs_unlink(char *name);
int vfs_rename(char *name, char *new_name);

/*links i/o*/
int vfs_readlink(char *name, char *pname);
int vfs_link(char *source_name, char *dest_name);
int vfs_follow_link(char *name, char *dest_name, int size);

/*point i/o*/
int vfs_mkpnt(char *name);
int vfs_rmpnt(char *name);
int vfs_read_pnt(char *name, dpoint *data);

/*misc i/o*/
int vfs_lock(char *name);
int vfs_unlock(char *name);

/*device i/o*/
int vfs_check_active(char *name);
int vfs_media_change(char *name);
int vfs_get_dev_load(char *name, void *data);
int vfs_get_dev_type(char *name, void *data);

/*permission i/o*/
int vfs_get_perm(char *name, perm_t *perm);
int vfs_set_perm(char *name, perm_t *perm);

#endif

