/************************************************************************************\
 * > ikernel/include/video/vga.h                                                    *
 * - Macros and defenitions for VGA functionality                                   *
 *                                                                                  *
 * Copyright(c) 2006 Dan Kruchinin (aka asgard)                                     *
 *                                                                                  *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     * 
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
\************************************************************************************/

#ifndef __VGA_H__
#define __VGA_H__

#include <klib/types.h>
#include <arch/types.h>

/* VGA memory base address */
#define VGA_TM_BASE 0xB8000 /* text mode */ 
#define VGA_VM_BASE 0xA0000 /* video mode */

/* VGA data register ports */
#define VDP_CRT    0x305 /* CRTC Controller Data Register (0x0B0 - mono; 0x0D0 - color) */
#define VDP_ATR_WR 0x3C0 /* Attribute Controller Data Write Register */
#define VDP_ATR_RD 0x3C1 /* Attribute Controller Data Read Register */
#define VDP_SDR    0x3C5 /* Sequencer Data Register */
#define VDP_DDR    0x3C9 /* DAC Data Register */
#define VDP_FCR_RD 0x3CA /* Feature Control Register */
#define VDP_MOR_WR 0x3C2 /* Miscellaneous Output Write Register */
#define VDP_MOR_RD 0x3CC /* Miscellaneous Output Read Register */
#define VDP_GCDR   0x3CF /* Graphics Controller Data Register */
#define VDP_ISR1   0x30A /* Input Status #1 Register (0x0B0 - mono; 0x0D0 - color) */

/* VGA address register ports */
#define VAP_CRT    0x304 /* CRTC Controller Address Register (0x0B0 - mono; 0x0D0 - color) */
#define VAP_ATR_WR 0x3C0 /* Attribute Address Write Register */
#define VAP_SAR    0x3C5 /* Sequencer Address Register */
#define VAP_DAR_RD 0x3C7 /* DAC State Register */
#define VAP_DAR_WR 0x3C8 /* DAC Address Write Mode Register */
#define VAP_GCAR   0x3CE /* Graphics Controller Address Register */

/* start CRTC indexes defenition */
#define NCRTCI 25   /* number of CRTC indexes */
#define HTR    0x00 /* Horizontal Total Register */
#define EHDR   0x01 /* End Horizontal Display Register */
#define SHBR   0x02 /* Start Horizontal Blanking Register */
#define EHBR   0x03 /* End Horizontal Blanking Register */
#define SHRR   0x04 /* Start Horizontal Retrace Register */
#define EHRR   0x05 /* End Horizontal Retrace Register */
#define VTR    0x06 /* Vertical Total Register */
#define OVR    0x07 /* Overflow Register */
#define PRSR   0x08 /* Preset Row Scan Register */
#define MSLR   0x09 /* Maximum Scan Line Register */
#define CSR    0x0A /* Cursor Start Register */
#define CER    0x0B /* Cursor End Register */
#define SAHR   0x0C /* Start Address High Register */
#define SALR   0x0D /* Start Address Low Register */
#define CLHR   0x0E /* Cursor Location High Register */
#define CLLR   0x0F /* Cursor Location Low Register */
#define VRSR   0x10 /* Vertical Retrace Start Register */
#define VRER   0x11 /* Vertical Retrace End Register */
#define VDER   0x12 /* Vertical Display End Register */
#define OFR    0x13 /* Offset Register */
#define ULR    0x14 /* Underline Location Register */
#define SVBR   0x15 /* Start Vertical Blanking Register */
#define EVBR   0x16 /* End Vertical Blanking */
#define CMCR   0x17 /* CRTC Mode Control Register */
#define LCR    0x18 /* Line Compare Register */
/* end of CRTC indexes defenition */

/* start Attribute indexes defenition */
#define NATRI 5    /* number of attribute indexes */
#define AMCR   0x10 /* Attribute Mode Control Register */
#define OCR    0x11 /* Overscan Color Register */
#define CPER   0x12 /* Color Plane Enable Register */
#define HPPR   0x13 /* Horizontal Pixel Panning Register */
#define CLSR   0x14 /* Color Select Register */
/* end of Attribute indexes defenition */

/* some major key codes */
#define BACKSPACE 0x08
#define TAB       0x09
#define SPACE     0x20

/* modes */
#define VGA_MONO  0x0B0 /* mono mode */
#define VGA_COLOR 0x0D0 /* color mode */

/* 16 major colors */
#define BLACK         0x00
#define BLUE          0x01
#define GREEN         0x02
#define CYAN          0x03
#define RED           0x04
#define MAGENTA       0x05
#define BROWN         0x06
#define LIGHR_GREY    0x07
#define DARK_GREY     0x08
#define LIGHT_BLUE    0x09
#define LIGHT_GREEN   0x0A
#define LIGHT_CYAN    0x0B
#define LIGHT_RED     0x0C
#define LIGHT_MAGENTA 0x0D
#define LIGHT_BROWN   0x0E
#define WHITE         0x0F

/* misc VGA flags */

/* CSRT register flgas */
#define VGA_CDISABLE 0x20
#define VGA_CENABLE  ((uint8_t)~0x20)

/* Attribute register flags */
#define VGA_ABLINK_ENABLE  0x08             /* blinking enable */
#define VGA_ABLINK_DISABLE ((uint8_t)~0x08) /* blinking disable */

/* misc flags for VGA functions */
#define VGA_CRSR_LOW    0x01  /* line cursor */
#define VGA_CRSR_HIGH   0x02  /* block cursor */
#define VGA_CRSR_BLIK   0x04  /* blinking cursor */

#define ENABLE  1
#define DISABLE 0

/* get reading or writing register address(mono or color depend on type) */
#define VGA_SET_MODE(addr, type) ((addr) + (type))

typedef struct vga_font {
  uint16_t wide;
  uint16_t high;
} vga_font;

typedef struct vga_dev {
  uint16_t* base;
  uint16_t cols;
  uint16_t rows;
  uint16_t x;
  uint16_t y;
  uint16_t fg;
  uint16_t bg;
  uint16_t es_char;
  vga_font font;
} vga_dev;

typedef struct vga_regs {
  uint8_t crtc[NCRTCI];
  uint8_t atr[NATRI];
} vga_regs;

/* function defenitoins */
void vga_init();
void vga_cls();
void vga_update_cursor();
void vga_putch(uint16_t);

void vga_scroll();
void vga_cursor(const bool);
void vga_set_cursor_attrs(const uint8_t);
void vga_blinking(const bool);
void vga_set_x(const uint16_t);
void vga_set_y(const uint16_t);
void vga_set_fg(const uint16_t);
void vga_set_bg(const uint16_t);
uint16_t vga_get_cols();
uint16_t vga_get_rows();
uint16_t vga_get_fg();
uint16_t vga_get_bg();
uint16_t vga_get_x();
uint16_t vga_get_y();

#endif /* __VGA_H__ */
