/************************************************************************************\
 * > ikernel/include/arch/cpu.h                                                     *
 * - CPU vectors functions and defenitions                                          *
 *                                                                                  *
 * Copyright(c) 2006 Dan Kruchinin (aka asgard)                                     *
 *                                                                                  *
 *                                                                                  *
 * This program is free software; you can redistribute it and/or                    *
 * modify it under the terms of the GNU General Public License                      *
 * as published by the Free Software Foundation; either version 2                   *
 * of the License, or (at your option) any later version.                           *
 *                                                                                  *
 * This program is distributed in the hope that it will be useful,                  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of                   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    *
 * GNU General Public License for more details.                                     * 
 *                                                                                  *
 * You should have received a copy of the GNU General Public License                *
 * along with this program; if not, write to the Free Software                      *
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.  *
\************************************************************************************/

#ifndef __CPU_H__
#define __CPU_H__

#include <arch/types.h>

#define ADDR_LIMIT 0xFFFFFFFF /* virtual address limit - 4G */
#define GDT_SIZE 6 /* number of GDT entrys */
#define IDT_SIZE 255 /* number of IDT entrys */

/* GDT descriptors */
#define GDES_NULL  0 /* NULL descriptor */
#define GDES_KTEXT 1 /* kernel text */
#define GDES_KDATA 2 /* kernel data */
#define GDES_UTEXT 3 /* user text */
#define GDES_UDATA 4 /* user data */
#define GDES_TSS   5 /* TSS segment */

#define STF_DATA (2 << 3)             /* data segment */
#define STF_CODE (3 << 3)             /* code segment */
#define STF_TSS 0x09                  /* TSS segment(execute only accessed) */
#define STF_PRESENT (1 << 7)          /* segment is present */
#define STF_PAGE_ACCESSED 1           /* page accessed */
#define STF_PAGE_WRITABLE (1 << 1)    /* page writable */
#define STF_PAGE_CONNFORMING (1 << 2) /* page conformed */
#define DPL_KERNEL (0 << 5)           /* kernel privelege level */
#define DPL_USER   (3 << 5)           /* user privelege level */
/* end of GDT access flags defenitions */

/* table segment descriptor strucure */
struct tdesc_ptr {
  uint16_t limit;
  uint32_t base;
} __attribute__ ((packed));

typedef struct tdesc_ptr tdesc_ptr;

/* GDT structure */

struct gdt_entry {
  unsigned limit       :16;
  unsigned base_low    :16;
  unsigned base_middle :8;
  unsigned access      :8;
  unsigned seg_limit   :4;
  unsigned available   :1;
  unsigned unused      :1;
  unsigned opsize      :1;
  unsigned granularity :1;
  unsigned base_high   :8;
} __attribute__ ((packed));

typedef struct gdt_entry gdt_entry;

/* IDT structure */

struct idt_entry {
  unsigned offset_low  :16;
  unsigned selector    :16;
  unsigned unused      :8;
  unsigned access      :8;
  unsigned offset_high :16;
} __attribute__ ((packed));

typedef struct idt_entry idt_entry;

/* TSS structure */

/* FIXME: add major architecture independet data types */
struct tss_selector {
  uint32_t ptl; /* previous task link */
  uint32_t esp0;
  uint32_t ss0;
  uint32_t esp1;
  uint32_t ss1;
  uint32_t esp2;
  uint32_t ss2;
  uint32_t cr3; /* PDBR */
  uint32_t eip;
  uint32_t eflags;
  uint32_t eax;
  uint32_t ecx;
  uint32_t edx;
  uint32_t ebx;
  uint32_t esp;
  uint32_t ebp;
  uint32_t esi;
  uint32_t edi;
  uint32_t es;
  uint32_t cs;
  uint32_t ss;
  uint32_t ds;
  uint32_t fs;
  uint32_t gs;
  uint32_t ldt_ss; /* LDT segment selector */
  uint32_t io_mba; /* I/O map base address */
} __attribute__ ((packed));

typedef struct tss_selector tss_selector;

/* main proccessor stricure(uniqie for each CPU) */
typedef struct cpu {
  int32_t cpu_id;       /* physical proccessor id */
  gdt_entry* gdt;   /* proccessor GDT table */
  tss_selector* tss; /* proccessor TSS segment */
  /* TODO: add other major fields */
} cpu;


extern cpu* cpus;

void init_cpu_vectors();

/* GDT functions */
void gdt_init();
void gdt_set_base(gdt_entry*, const uint32_t);
void gdt_set_limit(gdt_entry*, const uint32_t);
static void __gdt_fill(gdt_entry*, const uint32_t, const uint32_t, const uint8_t, const int32_t, const int32_t);

/* IDT functions */
void idt_init();
void idt_set_offset(idt_entry*, const uint32_t);

/* misc CPU functions */
void _get_cpu_type(int*);

#endif /* __CPU_H__ */
