/*
 * Virtual file system implementation.
 * Basic operations.
 * The part of the Jari microkernel.
 *
 * Written by Alfeiks Kaanoken <akaanoken@gmail.com>
 * (c) 2006
 *
 * This software is free software; you can redistribute it and/or                
 * modify it under the terms of the GNU Lesser General Public                   
 * License as published by the Free Software Foundation; either                 
 * version 2 of the License, or (at your option) any later version.             
 *                                                                              
 * This software is distributed in the hope that it will be useful,              
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU            
 * Lesser General Public License for more details.                              
 *                                                                              
 * You should have received a copy of the GNU Lesser General Public             
 * License along with this library; if not, write to the                        
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,                 
 * Boston, MA 02111-1307, USA. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <perm.h>
#include <vfs/vfs.h>
#include <vfs/types.h>
#include <kipc/errno.h>

/* TODO:
 * - cache functionality
 * - permission check
 */

int vfs_create(char *name,perm_t *p,char type)
{
  super_fnode *super=NULL;
  fnode *node=NULL;
  fnode *tnode=NULL;
  dpoint *parent=NULL;
  char *tn=name;
  char *root=NULL;
  int _s=0;

  if(!tn && !p)
    return FAIL;

  super=lookup_superfs_fnode(name);
  if(!super)
    return FAIL;
  if(!(node=(fnode *)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }
  if(type!=VFS_POINT) {
    node->type=type;
    node->name=name;
    node->lock=0;
    node->time=0; /*FIXME:*/
    node->perm=*p;
  } else {
    free(super);
    free(node);
    return FAIL;
  }
  super->def_func->put_fnode(node);
  /*now we're need to get a dpoint*/
  if(!(parent=(dpoint*)malloc(sizeof(struct __dpoint)))) {
    free(super);
    free(node);
    return FAIL;
  }
  else {
    parent->parent=NULL;
  }
  _s=strlen((const char *)super->mpoint);
  while(_s!=0) {
    _s--;
    *tn++;
  }
  root=tn;
  if(!(tnode=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(parent);
    return FAIL;
  }
  while(tn!='\0') {
    if(*tn==PNT_SEP) { /*so, we're get a point*/
      *tn='\0';
      tnode=super->def_func->get_fnode(root);
      tnode->def_func->read_pnt(tnode,NULL,parent);
      *tn=PNT_SEP;
    }

    *tn++;
  }
  while(*tn!=PNT_SEP)
    *tn--;

  node->def_func->create(node,tn,parent);
  free(tnode);
  free(super);
  free(parent);

  return SUCCESS;
}

int vfs_open(char *name)
{
  super_fnode *super=NULL;
  fnode *node=NULL;
  char *tn=name;
  char *root=NULL;
  int _s=0;

  if(!(super=lookup_superfs_fnode(name)) && !name)
    return FAIL;

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  _s=strlen((const char*)super->mpoint);
  while(_s!=0) {
    _s--;
    *tn++;
  }
  root=tn;

  node=super->def_func->get_fnode(root);
  super->def_func->read_fnode(node);

  node->def_func->open(node);

  free(super);
  free(node);

  return SUCCESS;
}

int vfs_close(char *name)
{
  super_fnode *super=NULL;
  fnode *node;
  char *tn=name;
  char *root=NULL;
  int _s=0;

  if(!(super=lookup_superfs_fnode(name)) && !name)
    return FAIL;

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  _s=strlen((const char*)super->mpoint);
  while(_s!=0) {
    _s--;
    *tn++;
  }
  root=tn;

  node=super->def_func->get_fnode(root);
  super->def_func->read_fnode(node);

  node->def_func->close(node);

  free(super);
  free(node);

  return SUCCESS;
}

int vfs_read(char *name,unsigned long size,void *data)
{
  super_fnode *super=NULL;
  fnode *node;
  char *tn=name;
  char *root=NULL;
  int _s=0;

  if(!(super=lookup_superfs_fnode(name)) && !name)
    return FAIL;

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  _s=strlen((const char*)super->mpoint);
  while(_s!=0) {
    _s--;
    *tn++;
  }
  root=tn;

  node=super->def_func->get_fnode(root);
  super->def_func->read_fnode(node);

  node->def_func->read(node,data,size);

  free(super);
  free(node);

  return SUCCESS;
}

int vfs_write(char *name, unsigned long size,void *data)
{
  super_fnode *super=NULL;
  fnode *node;
  char *tn=name;
  char *root=NULL;
  int _s=0;

  if(!(super=lookup_superfs_fnode(name)) && !name) {
    return FAIL;
  }

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  _s=strlen((const char*)super->mpoint);
  while(_s!=0) {
    _s--;
    *tn++;
  }
  root=tn;

  node=super->def_func->get_fnode(root);
  super->def_func->read_fnode(node);

  node->def_func->write(node,data,size);

  free(super);
  free(node);

  return SUCCESS;
}

int vfs_exec(char *name, void *data)
{
  super_fnode *super=NULL;
  fnode *node;
  char *tn=name;
  char *root=NULL;
  int _s=0;

  if(!(super=lookup_superfs_fnode(name)) && !name)
    return FAIL;

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  _s=strlen((const char*)super->mpoint);
  while(_s!=0) {
    _s--;
    *tn++;
  }
  root=tn;

  node=super->def_func->get_fnode(root);
  super->def_func->read_fnode(node);

  node->def_func->exec(node,data);

  free(super);
  free(node);

  return SUCCESS;
}

