/*
 * Virtual file system implementation.
 * File operations.
 * The part of the Jari microkernel.
 *
 * Written by Alfeiks Kaanoken <akaanoken@gmail.com>
 * (c) 2006
 *
 * This software is free software; you can redistribute it and/or                
 * modify it under the terms of the GNU Lesser General Public                   
 * License as published by the Free Software Foundation; either                 
 * version 2 of the License, or (at your option) any later version.             
 *                                                                              
 * This software is distributed in the hope that it will be useful,              
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU            
 * Lesser General Public License for more details.                              
 *                                                                              
 * You should have received a copy of the GNU Lesser General Public             
 * License along with this library; if not, write to the                        
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,                 
 * Boston, MA 02111-1307, USA. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <perm.h>
#include <vfs/vfs.h>
#include <vfs/types.h>
#include <kipc/errno.h>

unsigned long vfs_lseek(char *name, unsigned long seek)
{
  unsigned long seeks=0;
  super_fnode *super=NULL;
  fnode *node;
  char *tp=name;
  char *root=NULL;
  int _s=0;

  if(!(super=lookup_superfs_fnode(name)) && !name)
    return FAIL;

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  node=super->def_func->get_fnode(name);
  super->def_func->read_fnode(node);
  _s=strlen((const char*)super->mpoint);
  while(_s!=0)
    *tp++;
  root=tp;

  seeks=node->def_func->lseek(node,seek);

  free(super);
  free(node);

  return seeks;
}

int vfs_mmap(char *name,void *data)
{
  super_fnode *super=NULL;
  fnode *node;
  char *tp=name;
  char *root=NULL;
  int _s=0;

  if(!(super=lookup_superfs_fnode(name)) && !name)
    return FAIL;

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  node=super->def_func->get_fnode(name);
  super->def_func->read_fnode(node);
  _s=strlen((const char*)super->mpoint);
  while(_s!=0)
    *tp++;
  root=tp;

  node->def_func->mmap(node,data);

  free(super);
  free(node);

  return SUCCESS;
}

int vfs_unlink(char *name)
{
  super_fnode *super=NULL;
  fnode *node;
  char *tp=name;
  char *root=NULL;
  int _s=0;

  if(!(super=lookup_superfs_fnode(name)) && !name)
    return FAIL;

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  node=super->def_func->get_fnode(name);
  super->def_func->read_fnode(node);
  _s=strlen((const char*)super->mpoint);
  while(_s!=0)
    *tp++;
  root=tp;

  node->def_func->unlink(node);
  super->def_func->clear_fnode(node);
  node=super->def_func->unset_fnode(node);

  free(super);
  free(node);

  return SUCCESS;
}

int vfs_rename(char *name, char *new_name)
{
  super_fnode *super=NULL;
  fnode *node=NULL;
  fnode *tnode=NULL;
  char *tn=name;
  char *tnn=new_name;
  char *root=NULL;
  int _s=0;
  dpoint *point=NULL;

  if(!(super=lookup_superfs_fnode(name)) || !name || !new_name)
    return FAIL;

  if(!(node=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    return FAIL;
  }

  node=super->def_func->get_fnode(name);
  super->def_func->read_fnode(node);
  _s=strlen((const char*)super->mpoint);
  while(_s!=0)
    *tn++;
  root=tn;

  /*try to get the dpoint of the new name*/
  if(!(point=(dpoint*)malloc(sizeof(struct __dpoint)))) {
    free(super);
    free(node);
    return FAIL;
  } else {
    point->parent=NULL;
  }
  if(!(tnode=(fnode*)malloc(sizeof(struct __fnode)))) {
    free(super);
    free(node);
    free(point);
    return FAIL;
  }
  while(tnn!='\0') {
    if(*tnn==PNT_SEP) { /*so, we're get a point*/
      *tnn='\0';
      tnode=super->def_func->get_fnode(tnn);
      tnode->def_func->read_pnt(tnode,NULL,point);
      *tnn=PNT_SEP;
    }

    *tnn++;
  }
  while(*tnn!=PNT_SEP)
    *tnn--;

  node->def_func->rename(node,new_name,point);

  free(super);
  free(node);
  free(tnode);
  free(point);

  return SUCCESS;
}
