/*
 * Virtual file system implementation.
 * SuperFileSystem operations.
 * The part of the Jari microkernel.
 *
 * Written by Alfeiks Kaanoken <akaanoken@gmail.com>
 * (c) 2006
 *
 * This software is free software; you can redistribute it and/or                
 * modify it under the terms of the GNU Lesser General Public                   
 * License as published by the Free Software Foundation; either                 
 * version 2 of the License, or (at your option) any later version.             
 *                                                                              
 * This software is distributed in the hope that it will be useful,              
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU            
 * Lesser General Public License for more details.                              
 *                                                                              
 * You should have received a copy of the GNU Lesser General Public             
 * License along with this library; if not, write to the                        
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,                 
 * Boston, MA 02111-1307, USA. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <perm.h>
#include <vfs/vfs.h>
#include <vfs/types.h>
#include <kipc/errno.h>

static super_fs *sfs=NULL;

/*general function for calling from the microkernel*/
int init_vfs(void) 
{
  if(sfs!=NULL)
    return FAIL;
  else {
    sfs=(super_fs *)malloc(sizeof(struct __super_fs));
    if(!sfs)
      return FAIL;
    sfs->name=NULL;
    sfs->next=NULL;
  }

  return SUCCESS;
}
/*-----------------------*/

static super_fs *__unlink_super_fs(super_fs *p);
static super_fnode *__lookup_superfs_fnode(char *name,super_fnode *p);

/* misc function, just create an element for future adding to the
 * super_fses list
 */
super_fs *create_super_fs(int rw,int srv_id,char *name,char *mpoint,
			  super_fnode *(*read_fnode) (super_fnode *,void *)) 
{
  super_fs *n_fs=NULL;

  /*check what are we're adding*/
  if(!name && !mpoint && !read_fnode)
    return NULL;

  /*so, create on*/
  n_fs=(super_fs *)malloc(sizeof(struct __super_fs));
  if(!n_fs)
    return NULL;
  n_fs->rw=rw;
  n_fs->srv_id=srv_id;
  n_fs->name=name;
  n_fs->mpoint=mpoint;
  n_fs->read_fnode=read_fnode;
  n_fs->next=NULL;

  return n_fs;
}

/*add this*/
int add_super_fs(super_fs *n)
{
  super_fs *tp=sfs;
  super_fs *pp=NULL;

  if(!n || !tp)
    return FAIL;

  if(!tp->next && !tp->name) {
    tp->rw=n->rw;
    tp->srv_id=n->srv_id;
    tp->name=strcpy(tp->name,n->name);
    tp->mpoint=strcpy(tp->name,n->mpoint);
    tp->read_fnode=n->read_fnode;
    tp->next=NULL;
    n=__unlink_super_fs(n);
  } else {
    while(tp!=NULL) {
      pp=tp;
      tp=tp->next;
    }
    tp=(super_fs *)malloc(sizeof(struct __super_fs));
    pp->next=tp;
    tp->rw=n->rw;
    tp->srv_id=n->srv_id;
    tp->name=strcpy(tp->name,n->name);
    tp->mpoint=strcpy(tp->name,n->mpoint);
    tp->read_fnode=n->read_fnode;
    tp->next=NULL;
    n=__unlink_super_fs(n);
  }

  return SUCCESS;
}

int del_super_fs(char *mpoint)
{
  super_fs *tp=sfs;
  super_fs *pp=NULL;
  super_fs *d=NULL;
  int found=0;

  if(!tp && !mpoint)
    return FAIL;

  while(tp!=NULL) {
    pp=tp;
    if(!strcmp((const char*)mpoint,tp->mpoint)) {
      found+=1;
      d=tp;
      tp=tp->next;
      pp->next=tp;
      d=__unlink_super_fs(d);
      break;
    }
    tp=tp->next;
  }
  if(!found)
    return FAIL;
  if(found > 1)
    return del_super_fs(mpoint);

  return SUCCESS;
}

super_fnode *lookup_superfs_fnode(char *name)
{
  super_fnode *fn=NULL;

  fn=(super_fnode *)malloc(sizeof(struct __super_fnode));
  if(!fn)
    return NULL;

  fn=__lookup_superfs_fnode(name,fn);
  if(!fn->def_func) {
    free(fn);
    return NULL;
  }
  else
    return fn;
}

/*after each operations on the superfs we're need to call this function*/
int flush_superfs(void)
{
  super_fs *tp=sfs;
  super_fs *pp=NULL;
  super_fs *np=NULL;
  int cl=0,nl=0;
  int bound=0;

  if(!tp)
    return FAIL;
  if(!tp->next) /*if one do nothing, don't spend a time on it*/
    return SUCCESS;

  while(tp!=NULL) {
    np=tp->next;
    if(np) {
      cl=strlen((const char*)tp->mpoint);
      nl=strlen((const char*)np->mpoint);
      if(cl > nl) {
	bound++;
	pp=np;
	np->next=tp;
	if(pp)
	  tp->next=pp->next;
	else {
	  tp->next=NULL;
	  break;
	}
      }
    }
    tp=tp->next;
  }

  if(bound > 0)
    return flush_superfs();

  return SUCCESS;
}

/*internal functions*/
static super_fs *__unlink_super_fs(super_fs *p)
{
  if(!p)
    return p;
  free(p->name);
  free(p->mpoint);
  free(p);
  p=NULL;

  return p;
}

static super_fnode *__lookup_superfs_fnode(char *name,super_fnode *p)
{
  super_fs *tp=sfs;
  super_fs *j=NULL;
  int nlen=strlen((const char*)name);
  int flen=0;

  if(!p)
    return NULL;

  while(tp!=NULL) {
    flen=strlen((const char*)tp->mpoint);
    if((flen <= nlen) && !strncmp((const char*)tp->mpoint,(const char*)name, flen))
      j=tp;
    if(flen > nlen)
      break;
    tp=tp->next;
  }
  if(j)
    return j->read_fnode(p,(void *)name);
  else
    return p;
}
