/*
 * mkfs.mfs for linux.
 * Minimal file system is a Jari init file system.
 * This utility formatting space by the MFS.
 *
 * Originally written by Alfeiks Kaänoken <akaanoken@gmail.com>
 * (c) 2006
 *
 * This software is free software; you can redistribute it and/or                
 * modify it under the terms of the GNU Lesser General Public                   
 * License as published by the Free Software Foundation; either                 
 * version 2 of the License, or (at your option) any later version.             
 *                                                                              
 * This software is distributed in the hope that it will be useful,              
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU            
 * Lesser General Public License for more details.                              
 *                                                                              
 * You should have received a copy of the GNU Lesser General Public             
 * License along with this library; if not, write to the                        
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,                 
 * Boston, MA 02111-1307, USA. 
 *
 * This is a part of the Jari user-space development tools.
 * Jari Operating System&Software Development Kit.
 *
 */

/*
 * Version 0.0.1a - for use with files
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <mfs/mfs.h>
#include <perm.h>

#define MKFS_VERSION  "0.0.1a"

#define _DEBUG  1

static int block_size=0;
static mfs_head_t fs_head;

void _print_usage(void)
{
  fprintf(stdout,"mkfs.mfs usage:\n");
  fprintf(stdout,"mkfs.mfs -b <low|meduim|high> <file>\n");
  fprintf(stdout,"-b mean the block size, values can be:\n");
  fprintf(stdout,"low - 1024 bytes, medium - 2048 bytes, \n");
  fprintf(stdout,"high - 4096 bytes.\n");
  fprintf(stdout,"mkfs.mfs version %s, the part of the Jari OS&SDK.\n",MKFS_VERSION);
}

void _init_header(void)
{
  fs_head.revision[0]=MINOR_REV;
  fs_head.revision[1]=MAJOR_REV;
  sprintf(fs_head.version,"%s%s",MFS_NAME,MFS_VERSION);
  fs_head.system=SYS_LINUX;
  fs_head.block_size=block_size;
  fs_head.free_blocks=0;
  fs_head.total_data_blocks=0;
  fs_head.last_checked=0;
}

int main(int argc, char **argv)
{
  char **arg=argv;
  char *ar=NULL;
  char *fname=NULL;
  int fd=0;
  struct stat st;

  if(argc < 2) {
    _print_usage();
    return 0;
  }

  **arg++;
  ar=*arg;
  if(*ar=='-') {
    *ar++;
    if(*ar=='b') {
      **arg++;
      ar=*arg;
      if(!ar) {
	fprintf(stderr,"No option given for block size.\n");
	_print_usage();
	return 0;
      }
      if(!strcasecmp((const char*)ar,"low")) {
	block_size=LOW_SIZE;
      } else if (!strcasecmp((const char*)ar,"medium")) {
	block_size=MEDIUM_SIZE;
      } else if (!strcasecmp((const char*)ar,"high")) {
	block_size=HIGH_SIZE;
      } else {
	fprintf(stderr,"Unknown option for block size given \"%s\".\n",ar);
	_print_usage();
	return 0;
      }
      **arg++;
      fname=*arg;
      if(!fname) {
	fprintf(stderr,"No file name given.\n");
	_print_usage();
	return 0;
      }
    } else {
      fprintf(stderr,"Unknown option - \"%s\"\n",ar);
      _print_usage();
      return 0;
    }
  } else {
    fname=*arg;
    block_size=MEDIUM_SIZE;
  }

  _init_header(); /*init header with default values*/

  /*let's open the file*/
  fd=open(fname,O_RDWR|O_EXCL);
  if(fd==-1) {
    fprintf(stderr,"Error opening \"%s\".\n",fname);
    /*TODO: check for errors.*/
    return -1;
  } else {
    if(fstat(fd,&st)) {
      fprintf(stderr,"Error stat for file \"%s\".\n",fname);
      /*TODO: check for errors.*/
      return -1;
    }
    int blocks=(int)st.st_size/block_size;
    int regb=blocks*sizeof(struct __mfs_block_t);
    int total_data_block=blocks-(regb/block_size)-1;
    int total_free_block=blocks-((blocks/100)*10);
    int count=0;
    int r_bl=regb/block_size;
    int b_res=blocks-total_free_block;
    int b_st;
    mfs_block_t block;
    int per=0;
    int tg=1;

#ifdef _DEBUG
    printf("File system blocks storage: %d bytes.\n",regb);
    printf("Total blocks for data: %d. \n",total_data_block);
    printf("Total free blocks: %d. \n",total_free_block);
#endif /*_DEBUG*/

    fs_head.free_blocks=total_free_block;
    fs_head.total_data_blocks=total_data_block;
    fs_head.last_checked=st.st_atime;

    fprintf(stdout,"Writing FileSystem header.\n");
    fflush(stdout);
    write(fd,(const void*)&fs_head,sizeof(struct __mfs_head_t));
    lseek(fd,(off_t)block_size,SEEK_SET);

    printf("Writing block information ");
    fflush(stdout);
    per=blocks/10;
    while(count!=blocks) {
      if(count<=r_bl)
	b_st=B_SYSTEM;
      else if(count<=b_res)
	b_st=B_RESERVED;
      else
	b_st=B_EMPTY;

      block.block_addr+=count*block_size;
      block.state=b_st;
      block.fill_in=0;

      write(fd,(const void*)&block,sizeof(struct __mfs_block_t));
      lseek(fd,(off_t)sizeof(struct __mfs_block_t),SEEK_CUR);
      count++;
      if(count>per*tg) {
	tg++;
	printf(".");
	fflush(stdout);
      }
    }
    fflush(stdout);
    printf("\n");

  fprintf(stdout,"Total space for data: %ld bytes. \n",fs_head.free_blocks*block_size);
  fprintf(stdout,"Total space reserved for file system structure storage: %d bytes. \n",
	  b_res*block_size);

  }

  close(fd);

  return 0;
}

