/*
 * YacFS, Yet Another Contributed FileSystem.
 * The prefferred file system for the Jari Operating System.
 * checker.c: tool for testing yacfs file image.
 *
 * Originally written by Alfeiks Kaänoken <akaanoken@gmail.com>
 * (c) 2006
 *
 * This software is free software; you can redistribute it and/or                
 * modify it under the terms of the GNU Lesser General Public                   
 * License as published by the Free Software Foundation; either                 
 * version 2 of the License, or (at your option) any later version.             
 *                                                                              
 * This software is distributed in the hope that it will be useful,              
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU            
 * Lesser General Public License for more details.                              
 *                                                                              
 * You should have received a copy of the GNU Lesser General Public             
 * License along with this library; if not, write to the                        
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,                 
 * Boston, MA 02111-1307, USA. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#include <yacfs.h>

/*
 * First of all I need to check super blocks
 * One by one, and header ofcouse.
 */

yacfs_head_t header;

void _print_usage(void)
{
  fprintf(stdout,"Usage:\n"\
	  "checker <filename>.\n");
  return;
}

void _print_header(void)
{
  fprintf(stdout,"YacFS header info:\n"\
	  "Reserved percent: %d\n"     \
	  "Free inodes: %lld\n"	       \
	  "Busy inodes: %lld\n"	       \
	  "Space size: %lld bytes\n"   \
	  "Total blocks: %lld\n"       \
	  "Used blocks: %lld\n"	       \
	  "Flashed: %d\n"	       \
	  "Inode size: %d\n"	       \
	  "Block size: %d\n"	       \
	  "Inode density: %d\n",
	  header.reserved_percent,header.free_inodes,
	  header.busy_inodes,header.space_size,
	  header.blocks,header.used_blocks,
	  header.flashed,header.inode_size,
	  header.block_size,header.inode_per_blocks);
  
  fflush(stdout);

  return;
}

int _check_header(int fd)
{
  unsigned long magic=0;

  /*check magic*/
  read(fd,&magic,sizeof(YACFS_MAGIC));
  if(magic==(unsigned long)YACFS_MAGIC) {
    fprintf(stderr,"Magic: OK.\n");
    lseek(fd,sizeof(YACFS_MAGIC),SEEK_SET);
    read(fd,&header,sizeof(struct __yacfs_head_t));
    _print_header();
  } else {
    fprintf(stderr,"Incorrect magic number.\n");
    return -1;
  }

  return 0;
}

void _print_inode(yacfs_inode_t i)
{
  fprintf(stdout,"Inode %lld info:\n"\
	  "Offset: %p:%p\n"	     \
	  "State: %d\n"		     \
	  "Path hash: %p\n"	     \
	  "Full name hash: %p\n"     \
	  "Owner id: %d\n"	     \
	  "Group id: %d\n"	     \
	  "Node id: %d\n"	     \
	  "OGN: %d%d%d\n"	     \
	  "Type: %c\n"		     \
	  "Name lenght: %d\n"	     \
	  "Inode parent: %lld\n"     \
	  "Childs: %d\n"	     \
	  "Continios: %lld\n"	     \
	  "Affected blocks: %d\n",
	  i.inode_number,i.m_offset.primary,i.m_offset.secondary,
	  i.state,i.path_hash,i.full_name_hash,
	  i.owner_id,i.group_id,i.node_id,i.ogn[0],i.ogn[1],i.ogn[2],
	  i.type,i.name_len,i.inode_parent,i.child_num,
	  i.inode_continios,i.affected_blocks);

  return;
}

int _check_inodes(int fd)
{
  yacfs_inode_t inode;
  yacfs_block_t block;
  unsigned long super_size=header.inode_size;
  unsigned long long count=0;
  void *buf=NULL;
  void *t;
  char r='\0';
  int sfd;

  /*so,let's seek to the data*/
  lseek(fd,sizeof(YACFS_MAGIC)+sizeof(struct __yacfs_head_t),SEEK_SET);

  super_size+=header.inode_per_blocks*(sizeof(struct __yacfs_block_t)+header.block_size);
  fprintf(stdout,"Super size: %ld bytes\n",super_size);
  buf=calloc(1,super_size);
  if(!buf) {
    fprintf(stdout,"Error alocating space for buffer.\n");
    return -1;
  }
  t=buf;

  sfd=open("/dev/stdin",O_RDONLY|O_EXCL);

  while(count!=(header.free_inodes+header.busy_inodes)) {
    read(fd,buf,super_size);

    memcpy(&inode,buf,sizeof(struct __yacfs_inode_t));
    _print_inode(inode);

    fflush(stdout);
    while(r!='\n')
      read(sfd,&r,sizeof(char));

    count++;
    t=buf;
    r='\0';
  }

  free(buf);
  close(sfd);

  return 0;
}

int main(int argc, char **argv)
{
  int fd;
  char *fname=NULL;

  if(argc<2) {
    _print_usage();
    return 0;
  } else {
    **argv++;
    fname=*argv;
  }

  if((fd=open(fname,O_RDONLY|O_EXCL))==-1) {
    fprintf(stderr,"%s\n",strerror(errno));
    return 0;
  } else {
    if (_check_header(fd)==-1) {
      fprintf(stderr,"Error reading header.\n");
      return 0;
    }
    if(_check_inodes(fd)==-1) {
      fprintf(stderr,"Error reading inodes.\n");
      return 0;
    }
  }

  return 0;
}
