/*
 * YacFS, Yet Another Contributed FileSystem.
 * The prefferred file system for the Jari Operating System.
 * wrio_file.c : write/read info from file image of
 * the YacFS
 *
 * Originally written by Alfeiks Kaänoken <akaanoken@gmail.com>
 * (c) 2006
 *
 * This software is free software; you can redistribute it and/or                
 * modify it under the terms of the GNU Lesser General Public                   
 * License as published by the Free Software Foundation; either                 
 * version 2 of the License, or (at your option) any later version.             
 *                                                                              
 * This software is distributed in the hope that it will be useful,              
 * but WITHOUT ANY WARRANTY; without even the implied warranty of               
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU            
 * Lesser General Public License for more details.                              
 *                                                                              
 * You should have received a copy of the GNU Lesser General Public             
 * License along with this library; if not, write to the                        
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,                 
 * Boston, MA 02111-1307, USA. 
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>

#include <yacfs.h>
#include <yacfs_logic.h>

static int fd=-1;
static unsigned long super_size=0;
static yacfs_head_t ihead;
static unsigned long hs=0;

/*TODO: make a checks for read and write functions*/
/*TODO: make a size checks*/

int open_yacfs_image(char *fname)
{
  unsigned long magic=0;

  if((fd=open(fname,O_RDWR|O_EXCL))==-1) {
    fprintf(stderr,"%s\n",strerror(errno));
    return -1;
  }

  lseek(fd,0,SEEK_SET);
  read(fd,&magic,sizeof(YACFS_MAGIC));
  if(magic!=(unsigned long)YACFS_MAGIC) {
    fprintf(stderr,"File: %s is not an yacfs image.\nAborting.\n",fname);
    close(fd);
    return -1;
  }

  read(fd,&ihead,sizeof(struct __yacfs_head_t));
  super_size=ihead.inode_size;
  super_size+=ihead.inode_per_blocks*(ihead.block_size+sizeof(struct __yacfs_block_t));
  if(super_size==0) {
    fprintf(stderr,"Super size is invalid.\nAborting.\n");
    close(fd);
    return -1;
  }

  hs=sizeof(YACFS_MAGIC)+sizeof(struct __yacfs_head_t);

  return 0;
}

yacfs_head_t *get_yacfs_image_head(void)
{
  return &ihead;
}

static int __affect_yacfs_image_head(void) 
{
  lseek(fd,sizeof(YACFS_MAGIC),SEEK_SET);
  write(fd,&ihead,sizeof(struct __yacfs_head_t));

  return 0;
}

int _read_yacfs_image_inode_n(yacfs_inode_t *inode, unsigned long long n)
{
  unsigned long long sk=n*super_size;

  lseek(fd,sk+hs,SEEK_SET);
  read(fd,inode,sizeof(struct __yacfs_inode_t));

  return 0;
}

int _write_yacfs_image_inode_n(yacfs_inode_t *inode, unsigned long long n)
{
  unsigned long long sk=n*super_size;

  lseek(fd,sk+hs,SEEK_SET);
  write(fd,inode,sizeof(struct __yacfs_inode_t));

  return 0;
}

/*
 * I'm using copying to the memory for special purposes.
 * It's a not an ugly method.
 */
int read_yacfs_image_inode_n(yacfs_inode *inode,unsigned long long n)
{
  char *name=NULL;
  yacfs_inode_t *ch=NULL;
  yacfs_offset_t *blk=NULL;

  if(_read_yacfs_image_inode_n(&(inode->inode),n)!=0) {
    fprintf(stderr,"Error writing inode %lld.\n",n);
    return -1;
  }

  if(inode->inode.state!=YACFS_INODEFREE) {
    /*read name*/
    if(inode->inode.name_len!=0) {
      if((name=calloc(1,inode->inode.name_len+1))==NULL) {
	fprintf(stderr,"Error allocating memory.\n");
	return -1;
      }
      read(fd,name,inode->inode.name_len+1);
      inode->name=strcpy(inode->name,(const char*)name);
      free(name);
    }

    /*read childs*/
    if(inode->inode.child_num>0) {
      if((ch=calloc(1,inode->inode.child_num*sizeof(struct __yacfs_inode_t)))==NULL) {
	fprintf(stderr,"Error allocating memory.\n");
	return -1;
      }
      read(fd,ch,inode->inode.child_num*sizeof(struct __yacfs_inode_t));
      if((inode->childs=malloc(inode->inode.child_num*sizeof(struct __yacfs_inode_t)))==NULL) {
	fprintf(stderr,"Error allocating memory.\n");
	return -1;
      }
      memcpy(inode->childs,ch,inode->inode.child_num*sizeof(struct __yacfs_inode_t));
      free(ch);
    }

    /*read blocks*/
    if(inode->inode.affected_blocks>0) {
      if((blk=calloc(1,inode->inode.affected_blocks*sizeof(struct __yacfs_offset_t)))==NULL) {
	fprintf(stderr,"Error allocating memory.\n");
	return -1;
      }
      read(fd,blk,inode->inode.affected_blocks*sizeof(struct __yacfs_offset_t));

      inode->blocks=malloc(inode->inode.affected_blocks*sizeof(struct __yacfs_offset_t));
      if(inode->blocks==NULL) {
	fprintf(stderr,"Error allocating memory.\n");
	return -1;
      }
      memcpy(inode->blocks,blk,inode->inode.affected_blocks*sizeof(struct __yacfs_offset_t));
      free(blk);
    }
  }

  return 0;
}

int write_yacfs_image_inode_n(yacfs_inode *inode,unsigned long long n)
{
  if(inode->name==NULL || inode->inode.type=='\0') {
    fprintf(stderr,"Trying to write a null inode.\nAborting.\n");
    return -1;
  }

  if(_write_yacfs_image_inode_n(&(inode->inode),n)!=0) {
    fprintf(stderr,"Error writing inode %lld.\n",n);
    return -1;
  }

  /*write name*/
  write(fd,inode->name,inode->inode.name_len);

  /*write childs*/
  if(inode->inode.child_num>0) 
    write(fd,inode->childs,sizeof(struct __yacfs_inode_t)*inode->inode.child_num);

  /*write blocks*/
  if(inode->inode.affected_blocks>0) 
    write(fd,inode->blocks,sizeof(struct __yacfs_offset_t)*inode->inode.affected_blocks);
  
  /*  ihead.free_inodes--;
  ihead.busy_inodes++;

  __affect_yacfs_image_head();*/

  return 0;
}

static unsigned long long __cal_abs_addr(yacfs_offset_t o) 
{
  return (unsigned long long)(o.primary | ((unsigned long long)o.secondary << 32));
}

static yacfs_offset_t __cal_offset_addr(unsigned long long abs_off)
{
  yacfs_offset_t p;

  p.primary=abs_off&0xffffffff;
  p.secondary=abs_off >> 32;

  return p;
}

static int read_block_info(yacfs_block_t *blk,yacfs_offset_t off)
{
  unsigned long long abs_addr=0;

  abs_addr=__cal_abs_addr(off);

  /*TODO: check*/
  lseek(fd,abs_addr-sizeof(struct __yacfs_block_t),SEEK_SET);
  read(fd,blk,sizeof(struct __yacfs_block_t));

  return 0;
}

/* this function just puts data to the block addressed by yacfs_block_t and
 * putting there info.
 * TODO: make the work with the indirected blocks
 */
static int write_data_block(yacfs_block_t blk,void *data) 
{
  unsigned long long addr=__cal_abs_addr(blk.offset);
  /*so, put the head*/
  /*TODO: make a check*/
  lseek(fd,addr-sizeof(struct __yacfs_block_t),SEEK_SET);
  if(blk.fill_in==0)
    blk.state=YACFS_BLOCKFULL;
  else
    blk.state=YACFS_BLOCKDIRTY;
  write(fd,(const void*)&blk,sizeof(struct __yacfs_block_t));

  if(blk.fill_in!=0)
    write(fd,(const void*)data,blk.fill_in);
  else
    write(fd,(const void*)data,ihead.block_size);

  return 0;
}

static int read_data_block(yacfs_offset_t off,void *data)
{
  int fill_in=0;
  yacfs_block_t blk;

  if(read_block_info(&blk,off)!=0)
    return YACFS_READERR;
  fill_in=blk.fill_in;

  /*TODO: check*/
  lseek(fd,__cal_abs_addr(off),SEEK_SET);
  if(fill_in>0) {
    read(fd,data,fill_in);
  } else
    read(fd,data,ihead.block_size);

  return 0;
}

/*
 * So, I've need inode for the better allocating space,
 * it mean that when I trying to write data for inode 
 * the best method is the put it in the near blocks of the inode,
 * in the my terminology put it at the same super block.
 * TODO: make a feature with indirected blocks.
 */
static int lookup_free_data_blocks(unsigned long size,
				   yacfs_inode *inode,unsigned long *aff_blk)
{
  int fill_in=0;
  unsigned long blocks_measured=0,u=0;
  unsigned long long free_blocks=ihead.blocks-ihead.used_blocks;
  unsigned long long used_inode_block=0;
  yacfs_offset_t off;
  yacfs_block_t a_blk;
  yacfs_offset_t *g=NULL;
  int use_old=0;

  /*let's fragment data*/
  blocks_measured=size/ihead.block_size;
  if(size%ihead.block_size!=0) {
    fill_in=size%ihead.block_size;
    blocks_measured++;
  }

  /*first we're need to known is already we're have a data that associated with inode*/
  /*TODO: make a check*/
  if(inode->blocks!=NULL) {
    used_inode_block=inode->inode.affected_blocks;
    /*now for except a holes make some calculations*/
    g=inode->blocks;
    while(u!=used_inode_block-1) {
      u++;
      *g++;
    }
    off=*g;

    if(read_block_info(&a_blk,off)!=0) /*read error*/
      return YACFS_READERR;

    if(a_blk.fill_in>0) {
      fill_in+=a_blk.fill_in;
      if(fill_in>ihead.block_size) { /*oops*/
	fill_in=fill_in-(ihead.block_size-a_blk.fill_in);
      } else if(fill_in<ihead.block_size) {
	blocks_measured--;
      } else if(fill_in==ihead.block_size) {
	fill_in=0;
      }
      use_old=1;
      
    } else /*rare case*/
      use_old=0;

  if(blocks_measured>free_blocks) { /*no space*/
    return YACFS_NOSPACE;
  } 

    inode->blocks=realloc(inode->blocks,
			  (inode->inode.affected_blocks*sizeof(struct __yacfs_offset_t))+
			  (blocks_measured*sizeof(struct __yacfs_offset_t)));
    inode->blocks=realloc(inode->blk,
			  (inode->inode.affected_blocks*sizeof(struct __yacfs_block_t))+
			  (blocks_measured*sizeof(struct __yacfs_block_t)));
  } else { /*rare case*/
    inode->blocks=NULL;
    inode->blk=NULL;
    inode->blocks=malloc(blocks_measured*sizeof(struct __yacfs_offset_t));
    inode->blk=malloc(blocks_measured*sizeof(struct __yacfs_block_t));
    if(inode->blocks==NULL || inode->blk==NULL)
      return YACFS_MEMERR; /*no memory avialable*/
  }

  *aff_blk=blocks_measured;

  /*now we're need to fill the values and find free blocks*/
  /*try to put data in the same super block or fill it up,
   * if no, that try to fo it with the following inode, if not found,
   * use circle mechanism.
   */
  unsigned long long offy=sizeof(YACFS_MAGIC)+sizeof(struct __yacfs_head_t);
  yacfs_block_t rizy_blk;
  int circle=0;
  unsigned int jj=0;
  yacfs_block_t *bk=inode->blk;
  yacfs_offset_t *b=inode->blocks;
  unsigned long cur_i=inode->inode.inode_number;
  int i=blocks_measured;
  int sb=ihead.inode_per_blocks;

  if(use_old!=0) { /*step to the right inode*/
    jj=inode->inode.affected_blocks;
    while(jj!=0) {
      *b++; *bk++;
    }
  }

  offy+=inode->inode.inode_number*(ihead.inode_size+((ihead.block_size+sizeof(struct __yacfs_block_t))*ihead.inode_per_blocks));
  offy+=ihead.inode_size+sizeof(struct __yacfs_block_t);
  
  for(i=blocks_measured;i!=0;) {
    for(sb=ihead.inode_per_blocks;sb!=0;sb--) {
      if(read_block_info(&rizy_blk,__cal_offset_addr(offy))!=0) {
	if(!use_old) {
	  free(inode->blocks);
	  free(inode->blk);
	  inode->blocks=NULL;
	  inode->blk=NULL;
	} else {
	  inode->blocks=realloc(inode->blocks,
				inode->inode.affected_blocks*sizeof(struct __yacfs_offset_t));
	  inode->blk=realloc(inode->blk,
			     inode->inode.affected_blocks*sizeof(struct __yacfs_block_t));
	}
	return YACFS_READERR;
      }
      
      if(rizy_blk.state==YACFS_BLOCKFREE) { /*yep, found this*/
	*bk=rizy_blk; i--;
	*b=rizy_blk.offset;
	if(i==0)
	  (*bk).fill_in=fill_in;
	*bk++;*b++;
      }
      offy+=ihead.block_size+sizeof(struct __yacfs_block_t);
      if(i==0)
	goto end_off; /*we're found all the blocks*/
    }
    
    cur_i++;
    if(cur_i>(ihead.free_inodes+ihead.busy_inodes) && circle!=0) { /*no space*/
      if(!use_old) {
	free(inode->blocks);
	free(inode->blk);
	inode->blocks=NULL;
	inode->blk=NULL;
      } else {
	inode->blocks=realloc(inode->blocks,
			      inode->inode.affected_blocks*sizeof(struct __yacfs_offset_t));
	inode->blk=realloc(inode->blk,
			   inode->inode.affected_blocks*sizeof(struct __yacfs_block_t));
      }
      return YACFS_NOSPACE;
    } else if(cur_i>(ihead.free_inodes+ihead.busy_inodes) && circle==0) {
      circle++; /*try to find blocks before this super block*/
      offy=sizeof(YACFS_MAGIC)+sizeof(struct __yacfs_head_t);
      cur_i=0;
    }
    
    offy+=ihead.inode_size+sizeof(struct __yacfs_block_t);
    
  }


 end_off:
  if(use_old!=0)
    inode->inode.affected_blocks+=blocks_measured;
  else
    inode->inode.affected_blocks=blocks_measured;

  ihead.used_blocks+=blocks_measured;
  __affect_yacfs_image_head();

  return use_old;
}

/* absolute read function
 * simple
 */
int read_yacfs_image_data_b(yacfs_inode *inode,void *data,unsigned long size)
{
  int fill_in=0;
  unsigned long blk_tr=0;
  void *p=data;
  yacfs_offset_t *bl=inode->blocks;

  if(!size || !data) /*nothing to do*/
    return 0;
  if(!inode->blocks || !inode->blk) /*no data*/
    return 0;

  /*first of all try to determine how many blocks we're need to read*/
  blk_tr=size/ihead.block_size;
  fill_in=size%ihead.block_size;
  if(!blk_tr || fill_in!=0)
    blk_tr++;
  if(blk_tr>inode->inode.affected_blocks) { /*try to read more*/
    return YACFS_NOSPACE;
  }

  while(blk_tr!=0) {
    if(read_data_block(*bl,p)!=0) 
      return YACFS_READERR;
    blk_tr--; *bl++;
    p+=ihead.block_size;
  }

  return 0;
}

int write_yacfs_image_data_b(yacfs_inode *inode,void *data,unsigned long size)
{
  unsigned long ab=0;
  int y=0;
  void *p=data;
  yacfs_block_t *bb=NULL;

  if(!size || !data)
    return 0;

  y=lookup_free_data_blocks(size,inode,&ab);
  if(y < 0)
    return y;
  if(!y) { /*don't use old blocks*/
    bb=inode->blk;
    while(ab!=0) {
      if(write_data_block(*bb,p)!=0) {
	return YACFS_WRERR;
      }
      *bb++;
      p+=ihead.block_size;
      ab--;
    }
  } else if(y==1) {
    bb=inode->blk;
    /*TODO: check this*/
    lseek(fd,__cal_abs_addr((*bb).offset)+(*bb).fill_in,SEEK_SET);
    write(fd,p,ihead.block_size-(*bb).fill_in);
    p+=ihead.block_size-(*bb).fill_in;
    *bb++; ab--;
    while(ab!=0) {
      if(write_data_block(*bb,p)!=0) {
	return YACFS_WRERR;
      }
      *bb++;
      p+=ihead.block_size;
      ab--;
    }

  } else
    return YACFS_UNKNERR;

  return 0;
}

/* There functions for working with file marker
 * the read function cannot override the data,
 * but write is can.
 */
int read_yacfs_image_data(yacfs_inode *inode,void *data,unsigned long size,
			  unsigned long seek)
{
  return 0;
}

int write_yacfs_image_data(yacfs_inode *inode,void *data,unsigned long size,
			   unsigned long seek)
{
  return 0;
}
